@@ -24,9 +24,7 @@
  *	Peter Hutterer (peter.hutterer@redhat.com)
  */
 
-#ifdef HAVE_CONFIG_H
 #include "config.h"
-#endif
 
 #define _GNU_SOURCE 1
 #include "libwacomint.h"
@@ -53,27 +51,27 @@ libwacom_class_string_to_enum(const char *class)
 	if (class == NULL || *class == '\0')
 		return WCLASS_UNKNOWN;
 
-	if (streq(class, "Intuos3"))
+	if (g_str_equal(class, "Intuos3"))
 		return WCLASS_INTUOS3;
-	if (streq(class, "Intuos4"))
+	if (g_str_equal(class, "Intuos4"))
 		return WCLASS_INTUOS4;
-	if (streq(class, "Intuos5"))
+	if (g_str_equal(class, "Intuos5"))
 		return WCLASS_INTUOS5;
-	if (streq(class, "Cintiq"))
+	if (g_str_equal(class, "Cintiq"))
 		return WCLASS_CINTIQ;
-	if (streq(class, "Bamboo"))
+	if (g_str_equal(class, "Bamboo"))
 		return WCLASS_BAMBOO;
-	if (streq(class, "Graphire"))
+	if (g_str_equal(class, "Graphire"))
 		return WCLASS_GRAPHIRE;
-	if (streq(class, "Intuos"))
+	if (g_str_equal(class, "Intuos"))
 		return WCLASS_INTUOS;
-	if (streq(class, "Intuos2"))
+	if (g_str_equal(class, "Intuos2"))
 		return WCLASS_INTUOS2;
-	if (streq(class, "ISDV4"))
+	if (g_str_equal(class, "ISDV4"))
 		return WCLASS_ISDV4;
-	if (streq(class, "PenDisplay"))
+	if (g_str_equal(class, "PenDisplay"))
 		return WCLASS_PEN_DISPLAYS;
-	if (streq(class, "Remote"))
+	if (g_str_equal(class, "Remote"))
 		return WCLASS_REMOTE;
 
 	return WCLASS_UNKNOWN;
@@ -84,23 +82,23 @@ type_from_str (const char *type)
 {
 	if (type == NULL)
 		return WSTYLUS_UNKNOWN;
-	if (streq(type, "General"))
+	if (g_str_equal(type, "General"))
 		return WSTYLUS_GENERAL;
-	if (streq(type, "Inking"))
+	if (g_str_equal(type, "Inking"))
 		return WSTYLUS_INKING;
-	if (streq(type, "Airbrush"))
+	if (g_str_equal(type, "Airbrush"))
 		return WSTYLUS_AIRBRUSH;
-	if (streq(type, "Classic"))
+	if (g_str_equal(type, "Classic"))
 		return WSTYLUS_CLASSIC;
-	if (streq(type, "Marker"))
+	if (g_str_equal(type, "Marker"))
 		return WSTYLUS_MARKER;
-	if (streq(type, "Stroke"))
+	if (g_str_equal(type, "Stroke"))
 		return WSTYLUS_STROKE;
-	if (streq(type, "Puck"))
+	if (g_str_equal(type, "Puck"))
 		return WSTYLUS_PUCK;
-	if (streq(type, "3D"))
+	if (g_str_equal(type, "3D"))
 		return WSTYLUS_3D;
-	if (streq(type, "Mobile"))
+	if (g_str_equal(type, "Mobile"))
 		return WSTYLUS_MOBILE;
 	return WSTYLUS_UNKNOWN;
 }
@@ -110,11 +108,11 @@ eraser_type_from_str (const char *type)
 {
 	if (type == NULL)
 		return WACOM_ERASER_NONE;
-	if (streq(type, "None"))
+	if (g_str_equal(type, "None"))
 		return WACOM_ERASER_NONE;
-	if (streq(type, "Invert"))
+	if (g_str_equal(type, "Invert"))
 		return WACOM_ERASER_INVERT;
-	if (streq(type, "Button"))
+	if (g_str_equal(type, "Button"))
 		return WACOM_ERASER_BUTTON;
 	return WACOM_ERASER_UNKNOWN;
 }
@@ -122,13 +120,13 @@ eraser_type_from_str (const char *type)
 WacomBusType
 bus_from_str (const char *str)
 {
-	if (streq(str, "usb"))
+	if (g_str_equal(str, "usb"))
 		return WBUSTYPE_USB;
-	if (streq(str, "serial"))
+	if (g_str_equal(str, "serial"))
 		return WBUSTYPE_SERIAL;
-	if (streq(str, "bluetooth"))
+	if (g_str_equal(str, "bluetooth"))
 		return WBUSTYPE_BLUETOOTH;
-	if (streq(str, "i2c"))
+	if (g_str_equal(str, "i2c"))
 		return WBUSTYPE_I2C;
 	return WBUSTYPE_UNKNOWN;
 }
@@ -182,8 +180,8 @@ match_from_string(const char *str, WacomBusType *bus, int *vendor_id, int *produ
 	return TRUE;
 }
 
-static gboolean
-libwacom_matchstr_to_match(WacomDevice *device, const char *matchstr)
+static WacomMatch *
+libwacom_match_from_string(const char *matchstr)
 {
 	char *name = NULL;
 	int vendor_id, product_id;
@@ -191,24 +189,22 @@ libwacom_matchstr_to_match(WacomDevice *device, const char *matchstr)
 	WacomMatch *match;
 
 	if (matchstr == NULL)
-		return FALSE;
+		return NULL;
 
-	if (streq(matchstr, GENERIC_DEVICE_MATCH)) {
+	if (g_str_equal(matchstr, GENERIC_DEVICE_MATCH)) {
 		name = NULL;
 		bus = WBUSTYPE_UNKNOWN;
 		vendor_id = 0;
 		product_id = 0;
 	} else if (!match_from_string(matchstr, &bus, &vendor_id, &product_id, &name)) {
 		DBG("failed to match '%s' for product/vendor IDs. Skipping.\n", matchstr);
-		return FALSE;
+		return NULL;
 	}
 
 	match = libwacom_match_new(name, bus, vendor_id, product_id);
-	libwacom_add_match(device, match);
-	libwacom_match_unref(match);
-
 	free(name);
-	return TRUE;
+
+	return match;
 }
 
 static gboolean
@@ -268,24 +264,21 @@ libwacom_parse_stylus_keyfile(WacomDeviceDatabase *db, const char *path)
 		g_free (type);
 
 		string_list = g_key_file_get_string_list (keyfile, groups[i], "PairedStylusIds", NULL, NULL);
+		stylus->paired_ids = g_array_new (FALSE, FALSE, sizeof(int));
 		if (string_list) {
-			GArray *array;
 			guint j;
 
-			array = g_array_new (FALSE, FALSE, sizeof(int));
 			for (j = 0; string_list[j]; j++) {
 				int val;
 
 				if (safe_atoi_base (string_list[j], &val, 16)) {
-					g_array_append_val (array, val);
-					stylus->num_ids++;
+					g_array_append_val (stylus->paired_ids, val);
 				} else {
 					g_warning ("Stylus %s (%s) Ignoring invalid PairedId value\n", stylus->name, groups[i]);
 				}
 			}
 
 			g_strfreev (string_list);
-			stylus->paired_ids = (int *) g_array_free (array, FALSE);
 		}
 
 		stylus->has_lens = g_key_file_get_boolean(keyfile, groups[i], "HasLens", &error);
@@ -310,15 +303,15 @@ libwacom_parse_stylus_keyfile(WacomDeviceDatabase *db, const char *path)
 
 			for (j = 0; string_list[j]; j++) {
 				WacomAxisTypeFlags flag = WACOM_AXIS_TYPE_NONE;
-				if (streq(string_list[j], "Tilt")) {
+				if (g_str_equal(string_list[j], "Tilt")) {
 					flag = WACOM_AXIS_TYPE_TILT;
-				} else if (streq(string_list[j], "RotationZ")) {
+				} else if (g_str_equal(string_list[j], "RotationZ")) {
 					flag = WACOM_AXIS_TYPE_ROTATION_Z;
-				} else if (streq(string_list[j], "Distance")) {
+				} else if (g_str_equal(string_list[j], "Distance")) {
 					flag = WACOM_AXIS_TYPE_DISTANCE;
-				} else if (streq(string_list[j], "Pressure")) {
+				} else if (g_str_equal(string_list[j], "Pressure")) {
 					flag = WACOM_AXIS_TYPE_PRESSURE;
-				} else if (streq(string_list[j], "Slider")) {
+				} else if (g_str_equal(string_list[j], "Slider")) {
 					flag = WACOM_AXIS_TYPE_SLIDER;
 				} else {
 					g_warning ("Invalid axis %s for stylus ID %s\n",
@@ -377,7 +370,7 @@ libwacom_setup_paired_attributes(WacomDeviceDatabase *db)
 	}
 }
 
-struct {
+static const struct {
 	const char       *key;
 	WacomButtonFlags  flag;
 } options[] = {
@@ -392,7 +385,7 @@ struct {
 	{ "OLEDs", WACOM_BUTTON_OLED }
 };
 
-struct {
+static const struct {
 	const char       *key;
 	WacomStatusLEDs   value;
 } supported_leds[] = {
@@ -402,7 +395,7 @@ struct {
 	{ "Touchstrip2",	WACOM_STATUS_LED_TOUCHSTRIP2 }
 };
 
-struct {
+static const struct {
 	const char             *key;
 	WacomIntegrationFlags   value;
 } integration_flags[] = {
@@ -424,6 +417,7 @@ libwacom_parse_buttons_key(WacomDevice      *device,
 		return;
 	for (i = 0; vals[i] != NULL; i++) {
 		char val;
+		WacomButton *button;
 
 		val = *vals[i];
 		if (strlen (vals[i]) > 1 ||
@@ -432,85 +426,118 @@ libwacom_parse_buttons_key(WacomDevice      *device,
 			g_warning ("Ignoring value '%s' in key '%s'", vals[i], key);
 			continue;
 		}
-		val -= 'A';
-		device->buttons[(int) val] |= flag;
+
+		button = g_hash_table_lookup(device->buttons, GINT_TO_POINTER(val));
+		if (!button) {
+			button = g_new0(WacomButton, 1);
+			g_hash_table_insert(device->buttons, GINT_TO_POINTER(val), button);
+		}
+
+		button->flags |= flag;
 	}
 
 	g_strfreev (vals);
 }
 
+static void
+reset_code(gpointer key, gpointer value, gpointer user_data)
+{
+	WacomButton *button = value;
+	button->code = 0;
+}
+
 static inline bool
 set_button_codes_from_string(WacomDevice *device, char **strvals)
 {
-	gint i;
+	bool success = false;
 
 	assert(strvals);
 
-	for (i = 0; i < device->num_buttons; i++) {
-		gint val;
+	for (guint i = 0; i < g_hash_table_size(device->buttons); i++) {
+		char key = 'A' + i;
+		int code;
+		WacomButton *button = g_hash_table_lookup(device->buttons, GINT_TO_POINTER(key));
+
+		if (!button) {
+			g_error("%s: Button %c is not defined, ignoring all codes\n",
+				device->name, key);
+			goto out;
+		}
 
 		if (!strvals[i]) {
 			g_error ("%s: Missing EvdevCode for button %d, ignoring all codes\n",
 				 device->name, i);
-			return false;
+			goto out;
 		}
 
-		if (!safe_atoi_base (strvals[i], &val, 16) || val < BTN_MISC || val >= BTN_DIGI) {
-			g_warning ("%s: Invalid EvdevCode %s for button %d, ignoring all codes\n",
-				   device->name, strvals[i], i);
-			return false;
+		if (!safe_atoi_base (strvals[i], &code, 16) || code < BTN_MISC || code >= BTN_DIGI) {
+			g_warning ("%s: Invalid EvdevCode %s for button %c, ignoring all codes\n",
+				   device->name, strvals[i], key);
+			goto out;
 		}
-		device->button_codes[i] = val;
+		button->code = code;
 	}
+	success = true;
 
-	return true;
+out:
+	if (!success)
+		g_hash_table_foreach(device->buttons, reset_code, NULL);
+
+	return success;
 }
 
 static inline void
 set_button_codes_from_heuristics(WacomDevice *device)
 {
-	gint i;
-	for (i = 0; i < device->num_buttons; i++) {
+	for (char key = 'A'; key <= 'Z'; key++) {
+		int code = 0;
+		WacomButton *button;
+
+		button = g_hash_table_lookup(device->buttons, GINT_TO_POINTER(key));
+		if (!button)
+			continue;
+
 		if (device->cls == WCLASS_BAMBOO ||
 		    device->cls == WCLASS_GRAPHIRE) {
-			switch (i) {
-			case 0:
-				device->button_codes[i] = BTN_LEFT;
-				break;
-			case 1:
-				device->button_codes[i] = BTN_RIGHT;
-				break;
-			case 2:
-				device->button_codes[i] = BTN_FORWARD;
-				break;
-			case 3:
-				device->button_codes[i] = BTN_BACK;
-				break;
+			switch (key) {
+			case 'A': code = BTN_LEFT; break;
+			case 'B': code = BTN_RIGHT; break;
+			case 'C': code = BTN_FORWARD; break;
+			case 'D': code = BTN_BACK; break;
 			default:
-				device->button_codes[i] = 0;
 				break;
 			}
 		} else {
 			/* Assume traditional ExpressKey ordering */
-			switch (i) {
-			case 0 ... 9:
-				device->button_codes[i] = BTN_0 + i;
-				break;
-			case 10 ... 15:
-				device->button_codes[i] = BTN_A + (i-10);
-				break;
-			case 16:
-			case 17:
-				device->button_codes[i] = BTN_BASE + (i-16);
-				break;
+			switch (key) {
+			case 'A': code = BTN_0; break;
+			case 'B': code = BTN_1; break;
+			case 'C': code = BTN_2; break;
+			case 'D': code = BTN_3; break;
+			case 'E': code = BTN_4; break;
+			case 'F': code = BTN_5; break;
+			case 'G': code = BTN_6; break;
+			case 'H': code = BTN_7; break;
+			case 'I': code = BTN_8; break;
+			case 'J': code = BTN_9; break;
+			case 'K': code = BTN_A; break;
+			case 'L': code = BTN_B; break;
+			case 'M': code = BTN_C; break;
+			case 'N': code = BTN_X; break;
+			case 'O': code = BTN_Y; break;
+			case 'P': code = BTN_Z; break;
+			case 'Q': code = BTN_BASE; break;
+			case 'R': code = BTN_BASE2; break;
 			default:
-				device->button_codes[i] = 0;
 				break;
 			}
 		}
 
-		if (device->button_codes[i] == 0)
-			g_warning ("Unable to determine evdev code for button %d (%s)", i, device->name);
+		if (code == 0)
+			g_warning ("Unable to determine evdev code for button %c (%s)",
+				   key, device->name);
+
+		button->code = code;
 	}
 }
 
@@ -521,8 +548,7 @@ libwacom_parse_button_codes(WacomDevice *device,
 	char **vals;
 
 	vals = g_key_file_get_string_list(keyfile, BUTTONS_GROUP, "EvdevCodes", NULL, NULL);
-	if (!vals ||
-	    !set_button_codes_from_string(device, vals))
+	if (!vals || !set_button_codes_from_string(device, vals))
 		set_button_codes_from_heuristics(device);
 
 	g_strfreev (vals);
@@ -534,16 +560,21 @@ libwacom_parse_num_modes (WacomDevice      *device,
 			  const char       *key,
 			  WacomButtonFlags  flag)
 {
+	GHashTableIter iter;
 	int num;
-	int i;
+	gpointer k, v;
 
 	num = g_key_file_get_integer (keyfile, BUTTONS_GROUP, key, NULL);
 	if (num > 0)
 		return num;
-	for (i = 0; i < device->num_buttons; i++) {
-		if (device->buttons[i] & flag)
+
+	g_hash_table_iter_init(&iter, device->buttons);
+	while (g_hash_table_iter_next(&iter, &k, &v)) {
+		WacomButton *button = v;
+		if (button->flags & flag)
 			num++;
 	}
+
 	return num;
 }
 
@@ -578,17 +609,15 @@ libwacom_parse_styli_list(WacomDeviceDatabase *db, WacomDevice *device,
 	guint i;
 
 	array = g_array_new (FALSE, FALSE, sizeof(int));
-	device->num_styli = 0;
 	for (i = 0; ids[i]; i++) {
 		const char *id = ids[i];
 
-		if (strneq(id, "0x", 2)) {
+		if (g_str_has_prefix(id, "0x")) {
 			int int_value;
 			if (safe_atoi_base (ids[i], &int_value, 16)) {
 				g_array_append_val (array, int_value);
-				device->num_styli++;
 			}
-		} else if (strneq(id, "@", 1)) {
+		} else if (g_str_has_prefix(id, "@")) {
 			const char *group = &id[1];
 			GHashTableIter iter;
 			gpointer key, value;
@@ -596,9 +625,8 @@ libwacom_parse_styli_list(WacomDeviceDatabase *db, WacomDevice *device,
 			g_hash_table_iter_init(&iter, db->stylus_ht);
 			while (g_hash_table_iter_next (&iter, &key, &value)) {
 				WacomStylus *stylus = value;
-				if (stylus->group && streq(group, stylus->group)) {
+				if (stylus->group && g_str_equal(group, stylus->group)) {
 					g_array_append_val (array, stylus->id);
-					device->num_styli++;
 				}
 			}
 		} else {
@@ -608,7 +636,7 @@ libwacom_parse_styli_list(WacomDeviceDatabase *db, WacomDevice *device,
 	/* Using groups means we don't get the styli in ascending order.
 	   Sort it so the output is predictable */
 	g_array_sort(array, styli_id_sort);
-	device->supported_styli = (int *) g_array_free (array, FALSE);
+	device->styli = array;
 }
 
 static WacomDevice*
@@ -626,6 +654,7 @@ libwacom_parse_tablet_keyfile(WacomDeviceDatabase *db,
 	char *paired;
 	char **string_list;
 	bool success = FALSE;
+	int num_buttons;
 
 	keyfile = g_key_file_new();
 
@@ -639,6 +668,7 @@ libwacom_parse_tablet_keyfile(WacomDeviceDatabase *db,
 
 	device = g_new0 (WacomDevice, 1);
 	device->refcnt = 1;
+	device->matches = g_array_new(TRUE, TRUE, sizeof(WacomMatch*));
 
 	string_list = g_key_file_get_string_list(keyfile, DEVICE_GROUP, "DeviceMatch", NULL, NULL);
 	if (!string_list) {
@@ -647,23 +677,23 @@ libwacom_parse_tablet_keyfile(WacomDeviceDatabase *db,
 	} else {
 		guint i;
 		guint nmatches = 0;
-		guint first_valid_match = 0;
 		for (i = 0; string_list[i]; i++) {
-			if (libwacom_matchstr_to_match (device, string_list[i]))
-				nmatches++;
-			if (nmatches == 1)
-				first_valid_match = i;
-		}
-		if (nmatches == 0) {
-			DBG("failed to match '%s' for product/vendor IDs in '%s'\n", string_list[i], path);
-			g_strfreev (string_list);
-			goto out;
-		}
-		if (nmatches > 1) {
+			WacomMatch *m = libwacom_match_from_string(string_list[i]);
+			if (!m) {
+				DBG("'%s' is an invalid DeviceMatch in '%s'\n",
+				    string_list[i], path);
+				continue;
+			}
+			libwacom_add_match(device, m);
+			nmatches++;
 			/* set default to first entry */
-			libwacom_matchstr_to_match(device, string_list[first_valid_match]);
+			if (nmatches == 1)
+				libwacom_set_default_match(device, m);
+			libwacom_match_unref(m);
 		}
 		g_strfreev (string_list);
+		if (nmatches == 0)
+			goto out;
 	}
 
 	paired = g_key_file_get_string(keyfile, DEVICE_GROUP, "PairedID", NULL);
@@ -693,7 +723,7 @@ libwacom_parse_tablet_keyfile(WacomDeviceDatabase *db,
 		for (i = 0; string_list[i]; i++) {
 			found = FALSE;
 			for (n = 0; n < G_N_ELEMENTS (integration_flags); n++) {
-				if (streq(string_list[i], integration_flags[n].key)) {
+				if (g_str_equal(string_list[i], integration_flags[n].key)) {
 					device->integration_flags |= integration_flags[n].value;
 					found = TRUE;
 					break;
@@ -721,10 +751,11 @@ libwacom_parse_tablet_keyfile(WacomDeviceDatabase *db,
 		libwacom_parse_styli_list(db, device, string_list);
 		g_strfreev (string_list);
 	} else {
-		device->supported_styli = g_new (int, 2);
-		device->supported_styli[0] = WACOM_ERASER_FALLBACK_ID;
-		device->supported_styli[1] = WACOM_STYLUS_FALLBACK_ID;
-		device->num_styli = 2;
+		int fallback_eraser = WACOM_ERASER_FALLBACK_ID;
+		int fallback_stylus = WACOM_STYLUS_FALLBACK_ID;
+		device->styli = g_array_new(FALSE, FALSE, sizeof(int));
+		g_array_append_val(device->styli, fallback_eraser);
+		g_array_append_val(device->styli, fallback_stylus);
 	}
 
 	/* Features */
@@ -760,36 +791,33 @@ libwacom_parse_tablet_keyfile(WacomDeviceDatabase *db,
 		g_warning ("Tablet '%s' has touch switch but no touch tool. This is impossible", libwacom_get_match(device));
 
 	device->num_strips = g_key_file_get_integer(keyfile, FEATURES_GROUP, "NumStrips", NULL);
-	device->num_buttons = g_key_file_get_integer(keyfile, FEATURES_GROUP, "Buttons", &error);
-	if (device->num_buttons == 0 &&
+
+	num_buttons = g_key_file_get_integer(keyfile, FEATURES_GROUP, "Buttons", &error);
+	if (num_buttons == 0 &&
 	    g_error_matches (error, G_KEY_FILE_ERROR, G_KEY_FILE_ERROR_KEY_NOT_FOUND)) {
 		g_warning ("Tablet '%s' has no buttons defined, do something!", libwacom_get_match(device));
 		g_clear_error (&error);
 	}
-	if (device->num_buttons > 0) {
-		device->buttons = g_new0 (WacomButtonFlags, device->num_buttons);
-		device->button_codes = g_new0 (gint, device->num_buttons);
-		libwacom_parse_buttons(device, keyfile);
-	}
 
+	device->buttons = g_hash_table_new_full(g_direct_hash, g_direct_equal,
+						NULL, g_free);
+	if (num_buttons > 0)
+		libwacom_parse_buttons(device, keyfile);
+
+	device->status_leds = g_array_new (FALSE, FALSE, sizeof(WacomStatusLEDs));
 	string_list = g_key_file_get_string_list(keyfile, FEATURES_GROUP, "StatusLEDs", NULL, NULL);
 	if (string_list) {
-		GArray *array;
 		guint i, n;
 
-		array = g_array_new (FALSE, FALSE, sizeof(WacomStatusLEDs));
-		device->num_leds = 0;
 		for (i = 0; string_list[i]; i++) {
 			for (n = 0; n < G_N_ELEMENTS (supported_leds); n++) {
-				if (streq(string_list[i], supported_leds[n].key)) {
-					g_array_append_val (array, supported_leds[n].value);
-					device->num_leds++;
+				if (g_str_equal(string_list[i], supported_leds[n].key)) {
+					g_array_append_val (device->status_leds, supported_leds[n].value);
 					break;
 				}
 			}
 		}
 		g_strfreev (string_list);
-		device->status_leds = (WacomStatusLEDs *) g_array_free (array, FALSE);
 	}
 
 	success = TRUE;
@@ -807,100 +835,111 @@ out:
 	return device;
 }
 
-static int
-scandir_tablet_filter(const struct dirent *entry)
+static bool
+has_suffix(const char *name, const char *suffix)
 {
-	const char *name = entry->d_name;
-	int len, suffix_len;
+	size_t len = strlen(name);
+	size_t suffix_len = strlen(suffix);
 
 	if (!name || name[0] == '.')
 		return 0;
 
-	len = strlen(name);
-	suffix_len = strlen(TABLET_SUFFIX);
 	if (len <= suffix_len)
-		return 0;
+		return false;
 
-	return streq(&name[len - suffix_len], TABLET_SUFFIX);
+	return g_str_equal(&name[len - suffix_len], suffix);
 }
 
 static int
-scandir_stylus_filter(const struct dirent *entry)
+is_tablet_file(const struct dirent *entry)
 {
-	const char *name = entry->d_name;
-	int len, suffix_len;
+	return has_suffix(entry->d_name, TABLET_SUFFIX);
+}
 
-	if (!name || name[0] == '.')
-		return 0;
-
-	len = strlen(name);
-	suffix_len = strlen(STYLUS_SUFFIX);
-	if (len <= suffix_len)
-		return 0;
-
-	return streq(&name[len - suffix_len], STYLUS_SUFFIX);
+static int
+is_stylus_file(const struct dirent *entry)
+{
+	return has_suffix(entry->d_name, STYLUS_SUFFIX);
 }
 
 static bool
 load_tablet_files(WacomDeviceDatabase *db, const char *datadir)
 {
-    int n, nfiles;
-    struct dirent **files;
-    bool success = false;
+	DIR *dir;
+	struct dirent *file;
+	bool success = false;
+	GHashTable *keyset = NULL;
 
-    db->device_ht = g_hash_table_new_full (g_str_hash,
-					   g_str_equal,
-					   g_free,
-					   (GDestroyNotify) libwacom_destroy);
+	dir = opendir(datadir);
+	if (!dir)
+		return errno == ENOENT; /* non-existing directory is ok */
 
-    n = scandir(datadir, &files, scandir_tablet_filter, alphasort);
-    if (n <= 0)
-	    return false;
+	/* A set of all matches for duplicate detection. We allow duplicates
+	 * across data directories, but we don't allow for duplicates
+	 * within the same data directory.
+	 */
+	keyset = g_hash_table_new_full (g_str_hash, g_str_equal, free, NULL);
+	if (!keyset)
+		goto out;
 
-    nfiles = n;
-    while(n--) {
-	    WacomDevice *d;
-	    const WacomMatch **matches, **match;
+	while ((file = readdir(dir))) {
+		WacomDevice *d;
+		guint idx = 0;
 
-	    d = libwacom_parse_tablet_keyfile(db, datadir, files[n]->d_name);
+		if (!is_tablet_file(file))
+			continue;
 
-	    if (!d)
-		    continue;
+		d = libwacom_parse_tablet_keyfile(db, datadir, file->d_name);
+		if (!d)
+			continue;
 
-	    matches = libwacom_get_matches(d);
-	    if (!matches || !*matches) {
-		    g_critical("Device '%s' has no matches defined\n",
-			       libwacom_get_name(d));
-		    libwacom_destroy(d);
-		    goto out;
-	    }
+		if (d->matches->len == 0) {
+			g_critical("Device '%s' has no matches defined\n",
+				   libwacom_get_name(d));
+			goto out;
+		}
 
-	    for (match = matches; *match; match++) {
-		    const char *matchstr;
-		    matchstr = libwacom_match_get_match_string(*match);
-		    /* no duplicate matches allowed */
-		    if (g_hash_table_lookup(db->device_ht, matchstr) != NULL) {
-			    g_critical("Duplicate match of '%s' on device '%s'.",
-					matchstr, libwacom_get_name(d));
-			    goto out;
-		    }
-		    g_hash_table_insert (db->device_ht, g_strdup (matchstr), d);
-		    libwacom_ref(d);
-	    }
-	    libwacom_unref(d);
-    }
+		/* Note: we may change the array while iterating over it */
+		while (idx < d->matches->len) {
+			WacomMatch *match = g_array_index(d->matches, WacomMatch*, idx);
+			const char *matchstr;
 
-    if (g_hash_table_size (db->device_ht) == 0)
-	    goto out;
+			matchstr = libwacom_match_get_match_string(match);
+			/* no duplicate matches allowed within the same
+			 * directory */
+			if (g_hash_table_contains(keyset, matchstr)) {
+				g_critical("Duplicate match of '%s' on device '%s'.",
+					   matchstr, libwacom_get_name(d));
+				goto out;
+			}
+			g_hash_table_add(keyset, g_strdup(matchstr));
 
-    success = true;
+			/* We already have an entry for this match in the database,
+			 * that takes precedence. Remove the current match
+			 * from the new tablet - that's fine because we
+			 * haven't exposed the tablet yet.
+			 */
+			if (g_hash_table_lookup(db->device_ht, matchstr)) {
+				libwacom_remove_match(d, match);
+				continue;
+			}
+
+			g_hash_table_insert(db->device_ht,
+					    g_strdup (matchstr),
+					    d);
+			libwacom_ref(d);
+			idx++;
+		}
+		libwacom_unref(d);
+	}
+
+	success = true;
 
 out:
-    while(nfiles--)
-	    free(files[nfiles]);
-    free(files);
-
-    return success;
+	if (keyset)
+		g_hash_table_destroy(keyset);
+	closedir(dir);
+	return success;
 }
 
 static void
@@ -912,64 +951,85 @@ stylus_destroy(void *data)
 static bool
 load_stylus_files(WacomDeviceDatabase *db, const char *datadir)
 {
-    int n, nfiles;
-    struct dirent **files;
-    bool success = false;
+	DIR *dir;
+	struct dirent *file;
 
-    n = scandir(datadir, &files, scandir_stylus_filter, alphasort);
-    if (n <= 0)
-	    return false;
+	dir = opendir(datadir);
+	if (!dir)
+		return errno == ENOENT; /* non-existing directory is ok */
 
-    db->stylus_ht = g_hash_table_new_full (g_direct_hash,
-					   g_direct_equal,
-					   NULL,
-					   (GDestroyNotify) stylus_destroy);
-    nfiles = n;
-    while(n--) {
-	    char *path;
+	while ((file = readdir(dir))) {
+		char *path;
 
-	    path = g_build_filename (datadir, files[n]->d_name, NULL);
-	    libwacom_parse_stylus_keyfile(db, path);
-	    g_free(path);
-    }
+		if (!is_stylus_file(file))
+			continue;
 
-    /* If we couldn't load _anything_ then something's wrong */
-    if (g_hash_table_size (db->stylus_ht) == 0)
-	    goto end;
+		path = g_build_filename (datadir, file->d_name, NULL);
+		libwacom_parse_stylus_keyfile(db, path);
+		g_free(path);
+	}
 
-    success = true;
+	closedir(dir);
 
-end:
-    while(nfiles--)
-	    free(files[nfiles]);
-    free(files);
-
-
-    return success;
+	return true;
 }
 
+static WacomDeviceDatabase *
+database_new_for_paths (size_t npaths, const char **datadirs)
+{
+	WacomDeviceDatabase *db;
+	size_t n;
+	const char **datadir;
+
+	db = g_new0 (WacomDeviceDatabase, 1);
+	db->device_ht = g_hash_table_new_full (g_str_hash,
+					       g_str_equal,
+					       g_free,
+					       (GDestroyNotify) libwacom_destroy);
+	db->stylus_ht = g_hash_table_new_full (g_direct_hash,
+					       g_direct_equal,
+					       NULL,
+					       (GDestroyNotify) stylus_destroy);
+
+	for (datadir = datadirs, n = npaths; n--; datadir++) {
+		if (!load_stylus_files(db, *datadir))
+			goto error;
+	}
+
+	for (datadir = datadirs, n = npaths; n--; datadir++) {
+		if (!load_tablet_files(db, *datadir))
+			goto error;
+	}
+
+	/* If we couldn't load _anything_ then something's wrong */
+	if (g_hash_table_size (db->stylus_ht) == 0 ||
+	    g_hash_table_size (db->device_ht) == 0)
+		goto error;
+
+	libwacom_setup_paired_attributes(db);
+
+	return db;
+
+error:
+	libwacom_database_destroy(db);
+	return NULL;
+}
 
 LIBWACOM_EXPORT WacomDeviceDatabase *
 libwacom_database_new_for_path (const char *datadir)
 {
-    WacomDeviceDatabase *db;
-
-    db = g_new0 (WacomDeviceDatabase, 1);
-
-    if (!load_stylus_files(db, datadir) || !load_tablet_files(db, datadir)) {
-	    libwacom_database_destroy(db);
-	    db = NULL;
-    } else {
-	    libwacom_setup_paired_attributes(db);
-    }
-
-    return db;
+	return database_new_for_paths(1, &datadir);
 }
 
 LIBWACOM_EXPORT WacomDeviceDatabase *
 libwacom_database_new (void)
 {
-	return libwacom_database_new_for_path (DATADIR);
+	const char *datadir[] = {
+		ETCDIR,
+		DATADIR,
+	};
+
+	return database_new_for_paths (2, datadir);
 }
 
 LIBWACOM_EXPORT void
@@ -992,35 +1052,56 @@ device_compare(gconstpointer pa, gconstpointer pb)
 	cmp = libwacom_get_vendor_id(a) - libwacom_get_vendor_id(b);
 	if (cmp == 0)
 		cmp = libwacom_get_product_id(a) - libwacom_get_product_id(b);
+	if (cmp == 0)
+		cmp = g_strcmp0(libwacom_get_name(a), libwacom_get_name(b));
 	return cmp;
 }
 
+static void
+ht_copy_key(gpointer key, gpointer value, gpointer user_data)
+{
+	g_hash_table_add((GHashTable*)user_data, value);
+}
+
 LIBWACOM_EXPORT WacomDevice**
 libwacom_list_devices_from_database(const WacomDeviceDatabase *db, WacomError *error)
 {
-	GList *cur, *devices;
+	GList *cur, *devices = NULL;
 	WacomDevice **list, **p;
+	GHashTable *ht = NULL;
 
 	if (!db) {
 		libwacom_error_set(error, WERROR_INVALID_DB, "db is NULL");
 		return NULL;
 	}
 
-	devices = g_hash_table_get_values (db->device_ht);
+	/* Devices may be present more than one in the device_ht, so let's
+	 * use a temporary hashtable like a set to filter duplicates */
+	ht = g_hash_table_new (g_direct_hash, g_direct_equal);
+	if (!ht)
+		goto error;
+	g_hash_table_foreach (db->device_ht, ht_copy_key, ht);
+
+	devices = g_hash_table_get_keys (ht);
 	list = calloc (g_list_length (devices) + 1, sizeof (WacomDevice *));
-	if (!list) {
-		libwacom_error_set(error, WERROR_BAD_ALLOC, "Memory allocation failed");
-		g_list_free (devices);
-		return NULL;
-	}
+	if (!list)
+		goto error;
 
 	devices = g_list_sort (devices, device_compare);
-
 	for (p = list, cur = devices; cur; cur = g_list_next (cur))
 		*p++ = (WacomDevice *) cur->data;
 	g_list_free (devices);
+	g_hash_table_destroy (ht);
 
 	return list;
+
+error:
+	libwacom_error_set(error, WERROR_BAD_ALLOC, "Memory allocation failed");
+	if (ht)
+		g_hash_table_destroy (ht);
+	if (devices)
+		g_list_free (devices);
+	return NULL;
 }
 
 /* vim: set noexpandtab tabstop=8 shiftwidth=8: */
